/*
 *  ValueRef.java
 *
 *  $Id: Transducer.java 13256 2010-12-04 15:35:05Z johann_p $
 */

package at.ofai.gate.japebackrefs;

import gate.Annotation;
import gate.AnnotationSet;
import gate.util.GateRuntimeException;
import gate.jape.constraint.*;
import gate.jape.*;
import java.util.Map;
import java.util.HashMap;


/**
 * A ConstraintPredicate that makes it possible to use value back-references
 * in JAPE rules.
 * <p>
 * To use this, add the full name of this class to the list for the
 * operators initialisation parameter or use the JAPEBackRefs PR
 * that comes with this plugin.
 * <p>
 * In the JAPE grammer, use the valueref operator like this:
 * <pre>{Token valueref {Token.string == "refname"}}</pre>
 * This will either assign the current value of the string feature to the
 * reference name "refname" or if a value is already assigned, check that
 * the value of the feature is identical to what is already assigned.
 * If the reference name does not have a value assigned yet, the match will
 * always succeed. If a value is already assigned, the match will only succeed
 * if the feature value is identical (equals) the value stored for the reference
 * name. A reference remains stored between rules, phases and even documents
 * until it is explicitly reset using the "!=" operator instead of "==" like
 * this: <pre>{Token valueref {Token.string != "refname"}}</pre> This will
 * first make sure that any previous value for "refname" is removed and then
 * assign the value for the feature string to the reference name.
 * <p>
 * The following example will match sequences of two tokens where the
 * string and pos features are identical between both:
 * <pre>{Token valueref {Token.string != "sref"}, Token valueref {Token.pos != "pref"}}
 * {Token valueref {Token.string != "sref"}, Token valueref {Token.pos != "pref"}}</pre>
 * <p>
 * Note: the annotation type in the constraint to the right of "valueref" is
 * ignored, only the annotation type to the left of "valueref" is used for
 * accessing the feature. However, you have to specify an annotation type to
 * make the JAPE syntax checker happy, and to minimize confusion, it is best
 * to use the same annotation type for both left and right of "valueref".
 *
 * @author Johann Petrak
 */
public class ValueRef implements ConstraintPredicate {
  protected AnnotationAccessor accessor;
  protected Object value;
  
  static Map<String,Object> valuerefs = new HashMap<String,Object>();
  
  protected String referenceName = "";
  protected String featureName = "";
  protected boolean resetReference = false;
 
  public ValueRef() {
  }
  public ValueRef(AnnotationAccessor accessor, Object value) {
    // TODO: when is this ever called?
    System.out.println("MyConstraint: two args constructor "+accessor+"/"+value);
  }

  public boolean matches(Annotation annot, AnnotationSet context) 
    throws GateRuntimeException
  {
    if(resetReference) {
      valuerefs.remove(referenceName);
    }
    if(!valuerefs.containsKey(referenceName)) {
      valuerefs.put(referenceName,annot.getFeatures().get(featureName));
      return true;
    } else {
      if(valuerefs.get(referenceName).equals(annot.getFeatures().get(featureName))) {
        return true;
      } else {
        return false;
      }
    }
  }
  public void setAccessor(AnnotationAccessor accessor) {
    this.accessor = accessor;
    //System.out.println("MyConstraint: setAccessor "+accessor);
  }
  public AnnotationAccessor getAccessor() {
    //System.out.println("MyConstraint: getAccessor");
    return this.accessor;
  }
  public void setValue(Object value) {
    this.value = value;
    if(!(value instanceof Constraint)) {
      throw new GateRuntimeException("Not a constraint: "+value);
    }
    Constraint constraint = (Constraint)value;
    if(constraint.isNegated()) {
      throw new GateRuntimeException("Constraint must not be negated: "+constraint);
    }
    if(constraint.getAttributeSeq().size() != 1) {
      throw new GateRuntimeException("Constraint must have one predicate: "+constraint);
    }
    String annType = constraint.getAnnotType();
    //System.out.println("Annotation type is "+annType);
    ConstraintPredicate pred = constraint.getAttributeSeq().get(0);
    String op = pred.getOperator();
    //System.out.println("Operator is "+op);
    if(op.equals("!=")) {
      resetReference = true;
    } else if(!op.equals("==")) {
      throw new GateRuntimeException("Constraint operator must be == or !=");
    }
    Object predObj = pred.getValue();
    if(!(predObj instanceof String)) {
      throw new GateRuntimeException("Value of predicate must be a string: "+predObj);
    }
    String predValue = (String)predObj;
    //System.out.println("Value is "+predValue);
    referenceName = predValue;
    AnnotationAccessor aa = pred.getAccessor();
    String key = aa.getKey().toString();
    //System.out.println("Key is "+key);
    featureName = key;
    //System.out.println("Initialized constraint: ref="+referenceName+", feature="+featureName);
  }
  public Object getValue() {
    //System.out.println("MyConstraint: getValue");
    return this.value;
  }
  public String getOperator() {
    //System.out.println("MyConstraing: getOperator");
    return "valueref";
  }
}
